from django.test import TestCase
from polls import models
from nap.serialiser import Serialiser
from nap.models import ModelSerialiser
from nap.fields import Field


class SerializerTest(TestCase):

    def test_declared_fields(self):
        class MySerialiser(Serialiser):
            test_field = Field()
            another_field = Field()

        serializer = MySerialiser()
        for f in ('test_field', 'another_field'):
            self.assertIn(f, serializer._fields.keys())


class ModelSerializerTest(TestCase):

    def test_model_fields(self):
        class PollSerialiser(ModelSerialiser):
            class Meta:
                model = models.Poll

        serializer = PollSerialiser()
        for f in ('id', 'question', 'pub_date'):
            self.assertIn(f, serializer._fields.keys())

    def test_model_fields_include(self):

        class PollSerialiser(ModelSerialiser):
            class Meta:
                model = models.Poll
                fields = ('question', 'pub_date')

        serializer = PollSerialiser()
        for f in ('question', 'pub_date'):
            self.assertIn(f, serializer._fields.keys())
        self.assertNotIn('id', serializer._fields.keys())

    def test_model_fields_exclude(self):

        class PollSerialiser(ModelSerialiser):
            class Meta:
                model = models.Poll
                exclude = ('id')

        serializer = PollSerialiser()
        for f in ('question', 'pub_date'):
            self.assertIn(f, serializer._fields.keys())
        self.assertNotIn('id', serializer._fields.keys())




from . import fields
from .meta import Meta
from .serialiser import MetaSerialiser, Serialiser
from .publisher import Publisher

from django.shortcuts import get_object_or_404

FIELD_MAP = {}
# Auto-construct the field map
for f in dir(fields):
    cls = getattr(fields, f)
    try:
        if issubclass(cls, fields.Field):
            FIELD_MAP[f] = cls
    except TypeError:
        pass

class MetaModelSerialiser(MetaSerialiser):
    def __new__(mcs, name, bases, attrs):

        new_class = super(MetaModelSerialiser, mcs).__new__(mcs, name, bases, attrs)

        include = getattr(new_class._meta, 'fields', [])
        exclude = getattr(new_class._meta, 'exclude', [])

        current_fields = new_class._fields.keys()

        try:
            model = new_class._meta.model
            for f in model._meta.fields:
                # If we've got one, skip...
                if f.name in current_fields:
                    continue

                # If we have a whitelist, and it's not in it, skip
                if include and not f.name in include:
                    continue

                # If it's blacklisted, skip
                if f.name in exclude:
                    continue

                kwargs = {
                    'default': f.default,
                }

                field_class = FIELD_MAP.get(f.__class__.__name__, fields.Field)
                new_class._fields[f.name] = field_class(**kwargs)
        except AttributeError:
            pass

        return new_class


class ModelSerialiser(Serialiser):
    __metaclass__ = MetaModelSerialiser

    def restore_object(self, obj, instance, **kwargs):
        if instance:
            for key, val in obj.items():
                setattr(instance, key, val)
        else:
            instance = self._meta.model(**obj)
        if kwargs.get('commit', True):
            instance.save()
        return instance


class ModelPublisher(Publisher):
    '''A Publisher with useful methods to publish Models'''

    @property
    def model(self):
        '''By default, we try to get the model from our serialiser'''
        # XXX Should this call get_serialiser?
        return self.serialiser._meta.model

    # Auto-build serialiser from model class?

    def get_object_list(self):
        return self.model.objects.all()

    def get_object(self, object_id):
        return get_object_or_404(self.get_object_list(), pk=object_id)



class ModelFormMixin(object):
    '''Provide writable models using form validation'''

    initial = {}
    form_class = None

    # Here we mimic the FormMixin from django
    def get_initial(self):
        """
        Returns the initial data to use for forms on this view.
        """
        return self.initial.copy()

    def get_form_class(self):
        """
        Returns the form class to use in this view
        """
        return self.form_class

    def get_form(self, form_class):
        """
        Returns an instance of the form to be used in this view.
        """
        return form_class(**self.get_form_kwargs())

    def get_form_kwargs(self, **kwargs):
        """
        Returns the keyword arguments for instantiating the form.
        """
        kwargs.setdefault('initial', self.get_initial())
        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return kwargs

    def list_post_default(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        form = self.get_form(form_class)

        if form.is_valid():
            obj = form.save()
            return self.render_single_object(obj)

        # return errors


    def object_put_default(self, request, object_id, *args, **kwargs):
        obj = self.get_object(object_id)
        form_class = self.get_form_class()
        form = self.get_form(form_class, instance=obj)

        if form.is_valid():
            obj = form.save()
            return self.render_single_object(obj)

        # return errors

    def object_delete_default(self, request, object_id, *args, **kwargs):
        obj = self.get_object(object_id)
        # XXX Some sort of verification?
        obj.delete()
        return http.NoContent()


def modelserialiser_factory(name, model, **kwargs):
    attrs = {
        'model': model,
    }
    try:
        attrs['fields'] = kwargs['fields']
    except KeyError:
        pass
    try:
        attrs['exclude'] = kwargs['exclude']
    except KeyError:
        pass

    meta = type('Meta', (object,), attrs)
    return type(ModelSerialiser)(name, (ModelSerialiser,), {'Meta': meta})

class ModelSerialiserField(fields.SerialiserField):
    def __init__(self, serialiser=None, model=None, *args, **kwargs):
        if serialiser is None:
            serialiser = modelserialiser_factory(model.__name__ + 'Serialiser', model, **kwargs)()
        super(ModelSerialiserField, self).__init__(serialiser, *args, **kwargs)

class ModelManySerialiserField(fields.ManySerialiserField):
    def __init__(self, serialiser=None, model=None, *args, **kwargs):
        if serialiser is None:
            serialiser = modelserialiser_factory(model.__name__ + 'Serialiser', model, **kwargs)()
        super(ModelManySerialiserField, self).__init__(serialiser, *args, **kwargs)





from .utils import digattr

from decimal import Decimal
from datetime import datetime

class Field(object):
    type_class = None

    def __init__(self, attribute=None, default=None, readonly=False,
        *args, **kwargs):
        self.attribute = attribute
        self.default = default
        self.readonly = readonly
        self.args = args
        self.kwargs = kwargs

    def _get_attrname(self, name):
        return self.attribute if self.attribute else name

    def reduce(self, value, **kwargs):
        return value
    def restore(self, value, **kwargs):
        if self.type_class is not None:
            return self.type_class(value)
        return value

    def deflate(self, name, obj, data, **kwargs):
        src = self._get_attrname(name)
        value = digattr(obj, src, self.default)
        if value is not None:
            value = self.reduce(value, **kwargs)
        data[name] = value

    def inflate(self, name, data, obj, **kwargs):
        if self.readonly:
            return
        dest = self._get_attrname(name)
        try:
            value = data[name]
            obj[dest] = self.restore(value, **kwargs)
        except KeyError:
            pass

class BooleanField(Field):
    type_class = bool

class IntegerField(Field):
    type_class = int

class DecimalField(Field):
    type_class = Decimal
    def reduce(self, value, **kwargs):
        return float(value)


class DateTimeField(Field):
    def reduce(self, value, **kwargs):
        return value.replace(microsecond=0).isoformat(' ')
    def restore(self, value, **kwargs):
        return datetime.strptime(value, '%Y-%m-%d %H:%M:%S')


class DateField(Field):
    def reduce(self, value, **kwargs):
        return value.isoformat()
    def restore(self, value, **kwargs):
        return datetime.strptime(value, '%Y-%m-%d').date()


class TimeField(Field):
    def reduce(self, value, **kwargs):
        return value.isoformat()
    def restore(self, value, **kwargs):
        return datetime.strptime(value, '%H:%M:%S').time()


class SerialiserField(Field):
    def __init__(self, *args, **kwargs):
        super(SerialiserField, self).__init__(*args, **kwargs)
        self.serialiser = kwargs['serialiser']

    def reduce(self, value, **kwargs):
        return self.serialiser.object_deflate(value)
    def restore(self, value, **kwargs):
        return self.serialiser.object_inflate(value)


class ManySerialiserField(Field):
    def __init__(self, *args, **kwargs):
        super(ManySerialiserField, self).__init__(*args, **kwargs)
        self.serialiser = kwargs['serialiser']

    def reduce(self, value, **kwargs):
        return self.serialiser.list_deflate(iter(value), **kwargs)
    def restore(self, value, **kwargs):
        return self.serialiser.list_inflate(value, **kwargs)





from . import fields
from .meta import Meta


class MetaSerialiser(type):
    def __new__(mcs, name, bases, attrs):
        # Inherited fields
        attrs['_fields'] = {}

        # Inherit from parents
        try:
            for base in bases[::-1]:
                parent_fields = getattr(base, '_fields', {})
                attrs['_fields'].update(parent_fields)
        except NameError:
            # Can't do this for Serialiser
            pass

        declared_fields = {}
        for field_name, field in attrs.items():
            if isinstance(field, fields.Field):
                declared_fields[field_name] = attrs.pop(field_name)

        attrs['_fields'].update(declared_fields)

        new_class = super(MetaSerialiser, mcs).__new__(mcs, name, bases, attrs)
        meta = getattr(new_class, 'Meta', None)
        new_class._meta = Meta(meta)

        return new_class


class Serialiser(object):
    __metaclass__ = MetaSerialiser

    def object_deflate(self, obj, **kwargs):
        data = {}
        for name, field in self._fields.items():
            field.deflate(name, obj=obj, data=data, **kwargs)
            method = getattr(self, 'deflate_%s' % name, None)
            if method is not None:
                data[name] = method(obj=obj, data=data, **kwargs)
        return data

    def list_deflate(self, obj_list, **kwargs):
        return [
            self.object_deflate(obj, **kwargs)
            for obj in iter(obj_list)
        ]

    def object_inflate(self, data, instance=None, **kwargs):
        obj = {}
        for name, field in self._fields.items():
            if field.readonly:
                continue
            method = getattr(self, 'inflate_%s' % name, None)
            if method is not None:
                obj[name] = method(data=data, obj=obj, instance=instance, **kwargs)
            else:
                field.inflate(name, data, obj, **kwargs)
        return self.restore_object(obj, instance=instance, **kwargs)

    def list_inflate(self, data_list, **kwargs):
        # XXX target object list?
        return [
            self.object_inflate(data, **kwargs)
            for data in data_list
        ]

    def restore_object(self, obj, **kwargs): # pragma: no cover
        raise NotImplementedError
